from __future__ import print_function

import signal
import subprocess
import sys
from threading import Thread

try:
    from Queue import Queue, Empty
except ImportError:
    from queue import Queue, Empty  # Python 3.x

from .colour import get_colours
from .printer import Printer

ON_POSIX = 'posix' in sys.builtin_module_names


class Process(subprocess.Popen):
    """

    A simple utility wrapper around subprocess.Popen that stores
    a number of attributes needed by Honcho.

    """
    def __init__(self, cmd, name=None, *args, **kwargs):
        self.name = name
        self.reader = None
        self.printer = None
        self.dead = False

        defaults = {
            'stdout': subprocess.PIPE,
            'stderr': subprocess.STDOUT,
            'shell': True,
            'bufsize': 1,
            'close_fds': ON_POSIX
        }
        defaults.update(kwargs)

        super(Process, self).__init__(cmd, *args, **defaults)


class ProcessManager(object):
    """

    Here's where the business happens. The ProcessManager multiplexes and
    pretty-prints the output from a number of Process objects, typically added
    using the add_process() method.

    Example:

        pm = ProcessManager()
        pm.add_process('name', 'ruby server.rb')
        pm.add_process('name', 'python worker.py')

        pm.loop()

    """
    def __init__(self):
        self.processes = []
        self.colours = get_colours()
        self.queue = Queue()
        self.system_printer = Printer(sys.stdout, name='system')
        self.returncode = None

        self._terminating = False

    def add_process(self, name, cmd):
        """

        Add a process to this manager instance:

        Arguments:

        name        - a human-readable identifier for the process
                      (e.g. 'worker'/'server')
        cmd         - the command-line used to run the process
                      (e.g. 'python run.py')

        """
        self.processes.append(Process(cmd, name=name))

    def loop(self):
        """

        Enter the main loop of the program. This will print the multiplexed
        output of all the processes in this ProcessManager to sys.stdout, and
        will block until all the processes have completed.

        If one process terminates, all the others will be terminated by
        Honcho, and loop() will return.

        Returns: the returncode of the first process to exit, or 130 if
        interrupted with Ctrl-C (SIGINT)

        """

        self._init_readers()
        self._init_printers()

        for proc in self.processes:
            print("started with pid {0}".format(proc.pid), file=proc.printer)

        while True:
            try:
                proc, line = self.queue.get(timeout=0.1)
            except Empty:
                pass
            except KeyboardInterrupt:
                print("SIGINT received", file=sys.stderr)
                self.returncode = 130
                self.terminate()
            else:
                print(line, end='', file=proc.printer)

            for proc in self.processes:
                if not proc.dead and proc.poll() is not None:
                    print('process terminated', file=proc.printer)
                    proc.dead = True

                    # Set the returncode of the ProcessManager instance if not
                    # already set.
                    if self.returncode is None:
                        self.returncode = proc.returncode

                    self.terminate()

            if not self._process_count() > 0:
                break

        while True:
            try:
                proc, line = self.queue.get(timeout=0.1)
            except Empty:
                break
            else:
                print(line, end='', file=proc.printer)

        return self.returncode

    def terminate(self):
        """

        Terminate all the child processes of this ProcessManager, bringing the
        loop() to an end.

        """
        if self._terminating:
            return False

        self._terminating = True

        print("sending SIGTERM to all processes", file=self.system_printer)
        for proc in self.processes:
            if proc.poll() is None:
                print("sending SIGTERM to pid {0:d}".format(proc.pid), file=self.system_printer)
                proc.terminate()

        def kill(signum, frame):
            # If anything is still alive, SIGKILL it
            for proc in self.processes:
                if proc.poll() is None:
                    print("sending SIGKILL to pid {0:d}".format(proc.pid), file=self.system_printer)
                    proc.kill()

        signal.signal(signal.SIGALRM, kill)
        signal.alarm(5)

    def _process_count(self):
        return [p.poll() for p in self.processes].count(None)

    def _init_readers(self):
        for proc in self.processes:
            t = Thread(target=_enqueue_output, args=(proc, self.queue))
            t.daemon = True  # thread dies with the program
            t.start()

    def _init_printers(self):
        width = max(len(p.name) for p in self.processes)
        width = max(width, len(self.system_printer.name))

        self.system_printer.width = width

        for proc in self.processes:
            proc.printer = Printer(sys.stdout,
                                   name=proc.name,
                                   colour=self.colours.next(),
                                   width=width)


def _enqueue_output(proc, queue):
    for line in iter(proc.stdout.readline, b''):
        if not line.endswith('\n'):
            line += '\n'
        queue.put((proc, line))
    proc.stdout.close()



import argparse
import logging
import os
import re
import sys
from collections import defaultdict

from honcho import __version__
from honcho.procfile import Procfile
from honcho.process import Process, ProcessManager

logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.INFO)
log = logging.getLogger(__name__)

process_manager = ProcessManager()


# option decorator
def option(*args, **kwargs):
    def _decorator(func):
        _option = (args, kwargs)
        if hasattr(func, 'options'):
            func.options.append(_option)
        else:
            func.options = [_option]
        return func
    return _decorator

# arg decorator
arg = option


class Commander(type):
    def __new__(cls, name, bases, attrs):
        subcommands = {}
        commands = attrs.get('commands', [])
        for command in commands:
            func = attrs.get(command, None)
            if func is not None:
                subcommand = {
                    'name': command,
                    'func': func,
                    'options': []
                }
                if hasattr(func, 'options'):
                    subcommand['options'] = func.options
                subcommands[command] = subcommand
        attrs['_subcommands'] = subcommands
        return type.__new__(cls, name, bases, attrs)


class CommandError(Exception):
    pass


class Honcho(object):
    "Manage Procfile-based applications"
    __metaclass__ = Commander

    name = 'honcho'
    version = __version__
    epilog = ''
    formatter_class = argparse.ArgumentDefaultsHelpFormatter

    subparser_title = 'tasks'
    subparser_help = ''
    subparser_formatter_class = formatter_class

    default = ['--help']
    commands = ['start', 'check', 'help', 'run']
    common = [
        option('-e', '--env', help='Environment file[,file]', default='.env'),
        option('-d', '--app-root', help='Procfile directory', default='.'),
        option('-f', '--procfile', help='Procfile path', default='Procfile'),
    ]

    def add_common(self, parser):
        "add all common options and arguments to the main parser"
        common_group = parser.add_argument_group('common arguments')
        for option in self.common:
            options = option(lambda: None).options
            if options:
                args, kwargs = options[0]
                common_group.add_argument(*args, **kwargs)

    def parse(self, argv=None):
        # the main parser
        parser = argparse.ArgumentParser(
            prog=self.name,
            formatter_class=self.formatter_class,
            description=self.__doc__,
            epilog=self.epilog,
        )
        parser.add_argument('-v', '--version', action='version',
                            version='%(prog)s ' + self.version)
        self.add_common(parser)

        # then add the subparsers
        subparsers = parser.add_subparsers(
            title=self.subparser_title,
            help=self.subparser_help)

        for name, subcommand in sorted(self._subcommands.items()):
            subparser = subparsers.add_parser(subcommand['name'],
                                              help=subcommand['func'].__doc__,
                                              formatter_class=self.subparser_formatter_class)
            self.add_common(subparser)
            for args, kwargs in subcommand['options']:
                subparser.add_argument(*args, **kwargs)
            subparser.set_defaults(func=subcommand['func'])

        if not argv:
            argv = sys.argv[1:]
        if not argv:
            argv = self.default

        options = parser.parse_args(argv)

        try:
            options.func(self, options)
        except CommandError as e:
            if e.message:
                log.error(e.message)
            sys.exit(1)

    @arg('task', help='Task to show help for', nargs='?')
    def help(self, options):
        "Describe available tasks or one specific task"
        argv = ['--help']
        if options.task:
            argv.append(options.task)
        return self.parse(argv[::-1])

    def check(self, options):
        "Validate your application's Procfile"
        procfile = self.make_procfile(options.procfile)

        log.info('Valid procfile detected ({0})'.format(', '.join(procfile.commands)))

    @arg('command', nargs=argparse.REMAINDER, help='Command to run')
    def run(self, options):
        "Run a command using your application's environment"
        self.read_env(options)

        cmd = ' '.join(options.command)
        p = Process(cmd, stdout=sys.stdout, stderr=sys.stderr)
        p.wait()

    @option('-p', '--port', type=int, default=5000, metavar='N')
    @option('-c', '--concurrency', help='The number of each process type to run.', type=str, metavar='process=num,process=num')
    @arg('process', nargs='?', help='Name of process to start. All processes will be run if omitted.')
    def start(self, options):
        "Start the application (or a specific PROCESS)"
        self.read_env(options)
        procfile = self.make_procfile(options.procfile)

        port = int(os.environ.get('PORT', options.port))
        concurrency = self.parse_concurrency(options.concurrency)

        if options.process is not None:
            try:
                commands = {options.process: procfile.commands[options.process]}
            except KeyError:
                raise CommandError("Process type '{0}' does not exist in Procfile".format(options.process))
        else:
            commands = procfile.commands

        for name, cmd in commands.iteritems():
            for i in xrange(concurrency[name]):
                n = '{name}.{num}'.format(name=name, num=i + 1)
                os.environ['PORT'] = str(port + i)
                process_manager.add_process(n, cmd)
            port += 100

        sys.exit(process_manager.loop())

    def make_procfile(self, filename):
        try:
            with open(filename) as f:
                content = f.read()
        except IOError:
            raise CommandError('Procfile does not exist or is not a file')

        procfile = Procfile(content)

        if not procfile.commands:
            raise CommandError('No processes defined in Procfile')

        return procfile

    def read_env(self, args):
        app_root = args.app_root or os.path.dirname(args.procfile)
        files = [env.strip() for env in args.env.split(',')]
        for envfile in files:
            try:
                with open(os.path.join(app_root, envfile)) as f:
                    content = f.read()
                self.set_env(content)
            except IOError:
                pass

    def set_env(self, content):
        for line in content.splitlines():
            m1 = re.match(r'\A([A-Za-z_0-9]+)=(.*)\Z', line)
            if m1:
                key, val = m1.group(1), m1.group(2)

                m2 = re.match(r"\A'(.*)'\Z", val)
                if m2:
                    val = m2.group(1)

                m3 = re.match(r'\A"(.*)"\Z', val)
                if m3:
                    val = re.sub(r'\\(.)', r'\1', m3.group(1))

                os.environ[key] = val

    def parse_concurrency(self, desc):
        result = defaultdict(lambda: 1)
        if desc is None:
            return result
        for item in desc.split(','):
            key, concurrency = item.split('=', 1)
            result[key] = int(concurrency)
        return result


def main():
    app = Honcho()
    app.parse()

if __name__ == '__main__':
    main()



from datetime import datetime
import sys


class Printer(object):
    def __init__(self, output=sys.stdout, name='unknown', colour=None, width=0):
        self.output = output
        self.name = name
        self.colour = colour
        self.width = width

        self._write_prefix = True

    def write(self, *args, **kwargs):
        new_args = []

        for arg in args:
            lines = arg.split('\n')
            lines = [self._prefix() + l if l else l for l in lines]
            new_args.append('\n'.join(lines))

        self.output.write(*new_args, **kwargs)

    def _prefix(self):
        time = datetime.now().strftime('%H:%M:%S')
        name = self.name.ljust(self.width)
        prefix = '{time} {name} | '.format(time=time, name=name)
        if self.colour:
            return _colour_string(self.colour, prefix)
        else:
            return prefix


def _ansi(code):
    return '\033[{0}m'.format(code)


def _colour_string(colour, s):
    return '{0}{1}{2}'.format(_ansi(colour), s, _ansi(0))



import sys
from setuptools import setup, find_packages

from honcho import __version__

requirements = []
if sys.version_info[:2] < (2, 7):
    requirements.append('argparse')
    requirements.append('ordereddict')

setup(
    name='honcho',
    version=__version__,
    packages=find_packages(exclude=['test*']),

    # metadata for upload to PyPI
    author='Nick Stenning',
    author_email='nick@whiteink.com',
    url='https://github.com/nickstenning/honcho',
    description='Honcho: a python clone of Foreman. For managing Procfile-based applications.',
    license='MIT',
    keywords='sysadmin process procfile',

    install_requires=requirements,
    entry_points={
        'console_scripts': [
            'honcho=honcho.command:main'
        ]
    }
)
